/***************************************************************************

  gdocument.cpp

  (c) 2000-2005 Benoît Minisini <gambas@users.sourceforge.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#define __G_DOCUMENT_CPP

#include <qpainter.h>
#include <qscrollbar.h>
#include <qclipboard.h>
#include <qpixmap.h>
#include <qregexp.h>
#include <qapplication.h>
#include <qdragobject.h>
#include <qpopupmenu.h>
#include <qtimer.h>
#include <qdict.h>
#include <qcursor.h>
#include <qstyle.h>
#include <qcstring.h>
#include <qstring.h>

#include <ctype.h>

#include "main.h"
#include "gdocument.h"

static QColor defaultColors[GLine::NUM_STATE] =
{
  Qt::white,
  Qt::black,
  QColor(0x00, 0x80, 0xFF),
  Qt::blue,
  Qt::blue,
  Qt::black,
  Qt::red,
  Qt::magenta,
  Qt::gray,
  QColor(0xD0, 0x00, 0x00),
  QColor(0x00, 0x80, 0xFF),
  QColor(0x00, 0x80, 0xFF),
  QColor(0xC0, 0xC0, 0xFF),
  QColor(0xFF, 0xFF, 0x00),
  QColor(0xE8, 0xE8, 0xF8),
  Qt::red
};

/**---- GCommand -----------------------------------------------------------*/

class GCommand
{
public:
  enum Type
  {
    None, Begin, End, Insert, Delete, Indent, Unindent
  };

  virtual ~GCommand() { }
  virtual Type type() { return None; }
  virtual int nest() { return 0; }
  virtual void print() { }
  virtual bool merge(GCommand *) { return false; }
  virtual void process(GDocument *doc, bool undo) { }
};

class GBeginCommand: public GCommand
{
public:
  GBeginCommand() { }
  Type type() { return Begin; }
  void print() { qDebug("Begin"); }
  int nest() { return 1; }
};

class GEndCommand: public GCommand
{
public:
  GEndCommand() { }
  Type type() { return End; }
  void print() { qDebug("End"); }
  int nest() { return -1; }
  bool merge(GCommand *o) { return (o->type() == Begin); }
};

class GDeleteCommand: public GCommand
{
public:
  int x, y, x2, y2;
  QString str;

  GDeleteCommand(int y, int x, int y2, int x2, const QString & str)
  {
    this->x = x; this->y = y; this->x2 = x2; this->y2 = y2; this->str = str;
  }

  Type type() { return Delete; }
  void print() { qDebug("Delete: (%d %d)-(%d %d): '%s'", x, y, x2, y2, str.latin1()); }

  bool merge(GCommand *c)
  {
    if (c->type() != type())
      return false;

    GDeleteCommand *o = (GDeleteCommand *)c;

    if (x2 != o->x || y2 != o->y)
      return false;

    o->str.prepend(str);
    o->x = x;
    o->y = y;
    return true;
  }

  void process(GDocument *doc, bool undo)
  {
    if (undo)
      doc->insert(y, x, str);
    else
      doc->remove(y, x, y2, x2);
  }
};

class GInsertCommand: public GDeleteCommand
{
public:
  GInsertCommand(int y, int x, int y2, int x2, const QString & str): GDeleteCommand(y, x, y2, x2, str) {}
  Type type() { return Insert; }
  void print() { qDebug("Insert: (%d %d)-(%d %d): '%s'", x, y, x2, y2, str.latin1()); }

  bool merge(GCommand *c)
  {
    if (c->type() != type())
      return false;

    if (str[0] == '\n')
      return false;

    GInsertCommand *o = (GInsertCommand *)c;

    if (o->str[0] == '\n')
      return false;

    if (x != o->x2 || y != o->y2)
      return false;

    o->str += str;
    o->x2 = x2;
    o->y2 = y2;
    return true;
  }

  void process(GDocument *doc, bool undo)
  {
    GDeleteCommand::process(doc, !undo);
  }
};


/**---- GDocument -----------------------------------------------------------*/

#define FOR_EACH_VIEW(_view) \
  for (GEditor *_view = views->first(); _view ; _view =views->next())

GDocument::GDocument()
{
  oldCount = 0;
  oldMaxLength = 0;
  maxLength = 0;
  selector = NULL;
  blockUndo = false;
  tabWidth = 2;
  readOnly = false;
  highlightMode = None;
  baptismLimit = 0;

  lines = new QPtrList<GLine>;
  lines->setAutoDelete(true);

  views = new QPtrList<GEditor>;

  undoList.setAutoDelete(true);
  redoList.setAutoDelete(true);

  setText(QString("Gambas\nAlmost\nMeans\nBASIC!\n"));
}

GDocument::~GDocument()
{
  delete lines;
  delete views;
}

void GDocument::clear()
{
  uint i;

  clearUndo();
  lines->clear();
  lines->append(new GLine);
  baptismLimit = 0;

  updateViews();

  for (i = 0; i < views->count(); i++)
    views->at(i)->cursorGoto(0, 0, false);
}


void GDocument::reset()
{
  for (uint i = 0; i < lines->count(); i++)
    lines->at(i)->changed = false;

  updateViews();
}


QString GDocument::getText()
{
  QString tmp = "";

  if (lines->count())
  {
    for (uint i = 0; i < lines->count(); i++)
      colorize(i);

    for (uint i = 0; i < (lines->count() - 1); i++)
      tmp += lines->at(i)->s + '\n';

    tmp += lines->at(lines->count() - 1)->s;

    updateViews();
  }

  return tmp;
}

QString GDocument::getSelectedText() const
{
  QString tmp = "";
  int x1, y1, x2, y2;

  if (lines->count() && hasSelection())
  {
    getSelection(&y1, &x1, &y2, &x2);
    if (y1 == y2)
      tmp = lines->at(y1)->s.mid(x1, x2 - x1);
    else
    {
      tmp = lines->at(y1)->s.mid(x1) + '\n';
      for (int i = y1 + 1; i < y2; i++)
        tmp += lines->at(i)->s + '\n';

      tmp += lines->at(y2)->s.left(x2);
    }
  }

  return tmp;
}

int GDocument::getIndent(int y, bool *empty)
{
  int i;
  QString s = lines->at(y)->s;
  bool e = true;

  for (i = 0; i < (int)s.length(); i++)
  {
    if (!s[i].isSpace())
    {
      e = false;
      break;
    }
  }

  if (empty)
    *empty = e;

  return i;
}

void GDocument::insert(int y, int x, const QString & text)
{
  int pos = 0;
  int pos2;
  int xs = x, ys = y;
  GLine *l;
  int n = 1;
  QString rest;

  if (readOnly)
  {
    xAfter = x;
    yAfter = y;
    return;
  }

  for(;;)
  {
    pos2 = text.find('\n', pos);
    if (pos2 < 0)
      pos2 = text.length();

    //qDebug("insert '%s' at %d:%d", text.mid(pos, pos2 - pos).latin1(), y, x);

    l = lines->at(y);

    if (pos2 > pos)
    {
      l->s.insert(x, text.mid(pos, pos2 - pos));
      l->modified = l->changed = true;

      maxLength = QMAX(maxLength, (int)l->s.length());

      FOR_EACH_VIEW(v)
      {
        if (v->y == y && v->x >= x)
          v->x += pos2 - pos;
      }

      x += pos2 - pos;
    }

    pos = pos2 + 1;

    if (pos > (int)text.length())
      break;

    if (x < (int)l->s.length())
    {
      rest = l->s.mid(x);

      l->s.remove(x, rest.length());
      l->modified = l->changed = true;
    }

    FOR_EACH_VIEW(v)
    {
      if (v->y >= y)
        v->y++;
    }

    y++;

    lines->insert(y, new GLine());
    lines->at(y)->modified = lines->at(y)->changed = true;
    if (baptismLimit > y)
    	baptismLimit++;

    n = -1;
    x = 0;

  }

  if (n < 0 && rest.length())
  {
    l->s.insert(x, rest);
    l->modified = l->changed = true;

    maxLength = QMAX(maxLength, (int)l->s.length());
  }

  updateViews(ys, n);

  addUndo(new GInsertCommand(ys, xs, y, x, text));

  yAfter = y;
  xAfter = x;

  emit textChanged();
}

void GDocument::remove(int y1, int x1, int y2, int x2)
{
  GLine *l;
  QString text, rest;
  int y;

  yAfter = y1;
  xAfter = x1;

  if (readOnly)
    return;

  l = lines->at(y1);

  if (y1 == y2)
  {
    if (x2 >= x1)
    {
      text = l->s.mid(x1, x2 - x1);

      l->s.remove(x1, x2 - x1);
      l->modified = l->changed = true;

      FOR_EACH_VIEW(v)
      {
        if (v->y == y1 && v->x > x1)
          v->x = QMAX(x1, v->x - (x2 - x1));
      }

      updateViews(y1);
    }
  }
  else
  {
    text = l->s.mid(x1) + '\n';
    rest = lines->at(y2)->s.left(x2);

    l->s = l->s.left(x1) + lines->at(y2)->s.mid(x2);
    l->modified = l->changed = true;

    maxLength = QMAX(maxLength, (int)l->s.length());

    for (y = y1 + 1; y < y2; y++)
      text += lines->at(y)->s + '\n';
    text += rest;

    for (y = y1 + 1; y <= y2; y++)
    {
      lines->remove(y1 + 1);

			if (baptismLimit > (y1 + 1))
				baptismLimit--;
		}


    FOR_EACH_VIEW(v)
    {
      if (v->y > y1)
      {
        v->y = QMAX(y1, v->y - (y2 - y1));
        if (v->y == y1)
          v->x = x1;
      }
      else if (v->y == y1 && v->x > x1)
        v->x = x1;
    }

    updateViews(y1, -1);
  }

  addUndo(new GDeleteCommand(y1, x1, y2, x2, text));

  emit textChanged();
}

void GDocument::setText(const QString & text)
{
  bool oldReadOnly = readOnly;

  readOnly = false;
  blockUndo = true;

  clear();
  insert(0, 0, text);
  reset();

  blockUndo = false;
  readOnly = oldReadOnly;

  FOR_EACH_VIEW(v)
    v->cursorGoto(0, 0, false);
}

void GDocument::unsubscribe(GEditor *view)
{
  //qDebug("unsubscribe %p -> %p", view, this);
  views->removeRef(view);
	if (views->count() == 0)
		delete this;
}

void GDocument::subscribe(GEditor *view)
{
  //qDebug("subscribe %p -> %p", view, this);
  views->removeRef(view);
  views->append(view);
  view->setNumRows(lines->count());
  view->updateContents();
}

void GDocument::updateViews(int row, int count)
{
  uint i;

  if (lines->count() > oldCount)
  {
    oldCount = lines->count();
    FOR_EACH_VIEW(v)
    {
      v->setNumRows(oldCount);
      v->updateLength();
    }
  }

  if (maxLength != oldMaxLength)
  {
    oldMaxLength = maxLength;
    FOR_EACH_VIEW(v)
      v->updateLength();
  }

  if (row < 0)
  {
    row = 0;
    count = oldCount;
  }
  else if (count < 0)
  {
    count = oldCount - row;
  }

  count = QMIN((int)oldCount - row, count);

  FOR_EACH_VIEW(v)
  {
    for (i = row; i < (uint)(row + count); i++)
    {
      if (i >= lines->count())
        v->repaintCell(i, 0);
      else
        v->updateLine(i);
    }
  }

  if (lines->count() < oldCount)
  {
    oldCount = lines->count();
    FOR_EACH_VIEW(v)
    {
      v->setNumRows(oldCount);
      v->updateLength();
    }
  }

  FOR_EACH_VIEW(v)
    v->checkMatching();
}


void GDocument::getSelection(int *y1, int *x1, int *y2, int *x2) const
{
  if (!selector)
    return;

  if (ys2 > ys || (ys2 == ys && xs2 > xs))
  {
    *y1 = ys;
    *y2 = ys2;
    if (x1) *x1 = xs;
    if (x2) *x2 = xs2;
  }
  else
  {
    *y1 = ys2;
    *y2 = ys;
    if (x1) *x1 = xs2;
    if (x2) *x2 = xs;
  }
}

void GDocument::startSelection(GEditor *view, int y, int x)
{
  hideSelection();
  ys = y;
  xs = x;
  ys2 = ys;
  xs2 = xs;
  selector = view;
  updateViews(y);
}

void GDocument::endSelection(int y, int x)
{
  int y1a, y2a, y1b, y2b;

  getSelection(&y1a, NULL, &y2a, NULL);
  ys2 = y;
  xs2 = x;
  getSelection(&y1b, NULL, &y2b, NULL);

  if (y1a == y1b)
    updateViews(QMIN(y2a, y2b), QMAX(y2a, y2b) - QMIN(y2a, y2b) + 1);
  else if (y2a == y2b)
    updateViews(QMIN(y1a, y1b), QMAX(y1a, y1b) - QMIN(y1a, y1b) + 1);
  else
    updateViews(QMIN(y1a, y1b), QMAX(y2a, y2b) - QMIN(y1a, y1b) + 1);

  updateViews(y);
}

void GDocument::hideSelection()
{
  int y1, y2;

  if (!selector)
    return;

  getSelection(&y1, NULL, &y2, NULL);
  selector = NULL;
  updateViews(y1, y2 - y1 + 1);
}

void GDocument::eraseSelection()
{
  int y1, y2, x1, x2;

  if (!selector)
    return;

  getSelection(&y1, &x1, &y2, &x2);
  selector = NULL;
  /*x2--;
  if (x2 < 0)
  {
    y2--;
    x2 = lines->at(y2)->s.length() - 1;
  }*/
  remove(y1, x1, y2, x2);
}

void GDocument::copy() const
{
  QString text;

  if (!hasSelection())
    return;

  text = getSelectedText();

//#if defined(_WS_X11_)
//    disconnect(QApplication::clipboard(), SIGNAL(dataChanged()), this, 0);
//#endif
  QApplication::clipboard()->setText(text);
//#if defined(_WS_X11_)
//    connect(QApplication::clipboard(), SIGNAL(dataChanged()),
//            this, SLOT(clipboardChanged()));
//#endif
}


void GDocument::clearUndo()
{
  undoList.clear();
  redoList.clear();
}

void GDocument::addUndo(GCommand *c)
{
  if (blockUndo)
    return;

  //qDebug("addUndo");
  //c->print();

  if (!undoList.isEmpty())
  {
    if (c->merge(undoList.last()))
    {
      //qDebug("merge");
      delete c;
      return;
    }
  }

  undoList.append(c);

  if (!redoList.isEmpty())
  {
    redoList.clear();
    //emit redoAvailable(false);
  }
}

void GDocument::begin()
{
  addUndo(new GBeginCommand());
}

void GDocument::end()
{
  addUndo(new GEndCommand());
}

void GDocument::addRedo(GCommand * c)
{
  if (blockUndo)
    return;

  redoList.append(c);
}

bool GDocument::undo()
{
  int nest;

  if (undoList.isEmpty() || isReadOnly())
    return true;

  blockUndo = true;
  nest = 0;

  //qDebug("BEGIN UNDO");

  do
  {
    GCommand *c = undoList.take();
    if (!c)
      break;
    //c->print();
    c->process(this, true);
    nest += c->nest();
    //if (d->undoList.isEmpty())
    //  emit undoAvailable(false);
    redoList.append(c);
  }
  while (nest);

  //qDebug("END UNDO");

  blockUndo = false;
  return false;
}

bool GDocument::redo()
{
  int nest;

  if (redoList.isEmpty() || isReadOnly())
    return true;

  blockUndo = true;

  nest = 0;

  do
  {
    GCommand *c = redoList.take();
    if (!c)
      break;
    c->process(this, false);
    nest += c->nest();

    /*if (d->redoList.isEmpty())
      emit redoAvailable(false);
    if (d->undoList.isEmpty())
      emit undoAvailable(true);*/
    undoList.append(c);
  }
  while (nest);

  blockUndo = false;
  return false;
}

int GDocument::wordLeft(int y, int x, bool word)
{
  int xw;
  QString s = lines->at(y)->s;

  xw = x;
  for(;;)
  {
    if (xw == 0 || isWordChar(s[xw - 1]) ^ word)
      break;
    xw--;
  }

  return xw;
}

int GDocument::wordRight(int y, int x, bool word)
{
  int xw;
  QString s = lines->at(y)->s;

  xw = x;
  for(;;)
  {
    if (xw == (int)s.length() || isWordChar(s[xw]) ^ word)
      break;
    xw++;
  }

  return xw;
}

int GDocument::getLength() const
{
  uint i, len;

  if (lines->count())
    return 0;

  len = 0;

  for (i = 0; i < lines->count(); i++)
    len += lines->at(i)->s.length() + 1;

  return (len - 1);
}


QString GDocument::getLine(int y) const
{
  QString tmp = "";

  if (y >= 0 || y < (int)lines->count())
    tmp = lines->at(y)->s;

  return tmp;
}

void GDocument::setLine(int y, QString & str)
{
  if (y < 0 || y >= (int)lines->count())
    return;

	begin();
	if (lines->at(y)->s.length())
		remove(y, 0, y, lines->at(y)->s.length());
	if (str.length())
		insert(y, 0, str);
  end();
  updateViews(y);
}

bool GDocument::getLineFlag(int y, int f) const
{
  if (y >= 0 || y < (int)lines->count())
    return lines->at(y)->flag & (1 << f);
  else
    return false;
}

void GDocument::setLineFlag(int y, int f, bool b)
{
  if (y < 0 || y >= (int)lines->count())
    return;

  if (b)
    lines->at(y)->flag |= (1 << f);
  else
    lines->at(y)->flag &= ~(1 << f);

  updateViews(y);
}

int GDocument::convState(int state)
{
  switch(state)
  {
    case EVAL_TYPE_END: return GLine::Normal;
    case EVAL_TYPE_RESERVED: return GLine::Keyword;
    case EVAL_TYPE_IDENTIFIER: return GLine::Symbol;
    case EVAL_TYPE_NUMBER: return GLine::Number;
    case EVAL_TYPE_STRING: return GLine::String;
    case EVAL_TYPE_SUBR: return GLine::Subr;
    case EVAL_TYPE_COMMENT: return GLine::Comment;
    case EVAL_TYPE_OPERATOR: return GLine::Operator;
    case EVAL_TYPE_DATATYPE: return GLine::Datatype;
    case EVAL_TYPE_ERROR: return GLine::Error;
    default: return GLine::Normal;
  }
}

bool GDocument::highlightGambas(GEditor *, uint &state, QString &s, GHighlightArray &data, bool &proc)
{
  QCString cs;
  const char *src;
  EVAL_ANALYZE result;
  int i;
  uint ls;

  ls = s.length();
  cs = s.utf8();
  src = (const char *)cs;

  EVAL.Analyze(src, strlen(src), &result);

  data.resize(result.len);
  for (i = 0; i < result.len; i++)
  {
    //qDebug("state = %d -> %d  len = %d", result.color[i].state, QEditor::convState[result.color[i].state], result.color[i].len);
    data[i].state = convState(result.color[i].state);
    data[i].len = result.color[i].len;
  }

  s = QString::fromUtf8(result.str);
  GB.FreeString(&result.str);

  proc = result.proc;

  return qstrcmp((const char *)cs, result.str);
}

bool GDocument::highlightTest(GEditor *, uint &state, QString &s, GHighlightArray &data, bool &proc)
{
	QMemArray<uchar> d(s.length());
	uint i;
	int j;
	uint nstate = 0;

	if (s.length() == 0)
	{
		proc = false;
		return false;
	}

	for (i = 0; i < s.length(); i++)
	{
		nstate = state;

		if (state == GLine::Normal)
		{
			if (s[i] == '/' && i < (s.length() - 1) && s[i + 1] == '*')
				state = nstate = GLine::Comment;
		}
		else if (state == GLine::Comment)
		{
			if (s[i] == '/' && i > 0 && s[i - 1] == '*')
				nstate = GLine::Normal;
		}

		d[i] = state;
		state = nstate;
	}

  j = -1;
  data.resize(s.length());

  for (i = 0; i < s.length(); i++)
  {
  	if (i == 0 || d[i] != data[j].state)
  	{
  		j++;
  		data[j].state = d[i];
  		data[j].len = 1;
		}
		else
			data[j].len++;
  }

	if (j >= 0)
  	data.resize(j + 1);

  proc = false;

  return false; //qstrcmp((const char *)cs, result.str);
}

void GDocument::colorize(int y)
{
  GLine *l;
  bool modif;
  bool proc;
  QString old;
  uint state;
  int nupd = 0;

  if (highlightMode == None)
    return;

  if (y < 0 || y >= (int)lines->count())
    return;

	for(;;)
	{
		l = lines->at(y);

		if (!l->modified)
			break;

		//qDebug("colorize: %d", y);

		modif = false;

		if (y == 0)
			state = GLine::Normal;
		else
			state = lines->at(y - 1)->state;

		if (l->s.length())
		{
			old = l->s;
			modif = (*highlightCallback)(views->first(), state, l->s, l->highlight, proc);
			l->proc = proc;

			if (l->baptized)
			{
				if (old != l->s)
				{
					begin();
					addUndo(new GDeleteCommand(y, 0, y, old.length(), old));
					if (l->s.length())
						addUndo(new GInsertCommand(y, 0, y, l->s.length(), l->s));
					end();
				}
			}
			else
				l->baptized = true;
		}

		l->modified = false;

		if (l->state != state)
		{
			l->state = state;
			if (y < (numLines() - 1))
			{
				y++;
				lines->at(y)->modified = true;
				nupd++;
				continue;
			}
		}

		break;
	}

	if (nupd)
		updateViews(y - nupd + 1, nupd);
}


void GDocument::baptizeUntil(int y)
{
	bool busy = (y - baptismLimit) > 256;

	if (busy)
  	qApp->setOverrideCursor(Qt::waitCursor);

	while (baptismLimit <= y)
	{
		colorize(baptismLimit);
		baptismLimit++;
	}

	if (busy)
  	qApp->restoreOverrideCursor();
}

void GDocument::setHighlightMode(int mode, GHighlightCallback cb)
{
	int i;

	highlightMode = mode;

	if (mode == Gambas)
		highlightCallback = GDocument::highlightGambas;
	else
		highlightCallback = cb;

	for (i = 0; i < numLines(); i++)
	{
		lines->at(i)->modified = true;
		lines->at(i)->baptized = false;
	}

	baptismLimit = 0;

	updateViews();
}


/**---- GEditor -----------------------------------------------------------*/

QPixmap *GEditor::cache = 0;
int GEditor::count = 0;

GEditor::GEditor(QWidget *parent) : QGridView(parent, 0, WNoAutoErase)
{
  int i;

  if (count == 0)
    cache = new QPixmap();

  count++;

  setNumCols(1);
  setKeyCompression(true);
  setFocusPolicy(WheelFocus);
  setPaletteBackgroundColor(defaultColors[GLine::Background]);

  viewport()->setCursor(ibeamCursor);
  //viewport()->setWFlags(WRepaintNoErase);

  x = y = xx = 0;
  x1m = x2m = 0;
  ym = -1;
  cursor = false;
  margin = 0;
  lineNumberLength = 0;
  doc = 0;
  setDocument(NULL);

  for (i = 0; i < GLine::NUM_STATE; i++)
  {
    styles[i].color = defaultColors[i];
    styles[i].bold = i == GLine::Keyword;
    styles[i].italic = i == GLine::Comment;
    styles[i].underline = i == GLine::Error;
  }

  flags = 0;
  setFlag(ShowCurrentLine, true);
  setFont(QFont("monospace", QApplication::font().pointSize()));
  updateLength();

  blinkTimer = new QTimer(this);
  connect(blinkTimer, SIGNAL(timeout()), this, SLOT(blinkTimerTimeout()));
  //startBlink();

  scrollTimer = new QTimer(this);
  connect(scrollTimer, SIGNAL(timeout()), this, SLOT(scrollTimerTimeout()));

  connect(this, SIGNAL(contentsMoving(int, int)), this, SLOT(baptizeVisible()));
}

GEditor::~GEditor()
{
  doc->unsubscribe(this);
  count--;
  if (count == 0)
    delete cache;
}

void GEditor::setDocument(GDocument *d)
{
	if (doc)
	{
  	doc->unsubscribe(this);
		disconnect(doc, 0, this, 0);
	}

  doc = d;
  if (!doc)
    doc = new GDocument;

  doc->subscribe(this);
  connect(doc, SIGNAL(textChanged()), this, SLOT(docTextChanged()));
}

void GEditor::updateLength()
{
  QFontMetrics fm(font());
  charWidth = fm.width('m');

  updateMargin();

  setCellWidth(QMAX(visibleWidth(),  margin + charWidth * doc->getMaxLineLength() + 2));
  //qDebug("contentsWidth() = %d  cellWidth() = %d", contentsWidth(), cellWidth());
  setCellHeight(fm.lineSpacing());

  if (cache->width() < cellWidth() || cache->height() < cellHeight())
    cache->resize(cellWidth(), cellHeight());

	italicFont = font();
	italicFont.setItalic(true);
}

void GEditor::fontChange(const QFont &oldFont)
{
	QGridView::fontChange(oldFont);
  updateLength();
  updateContents();
}

void GEditor::paintText(QPainter &p, GLine *l, int x, int y, int xmin, int lmax)
{
  unsigned int i;
  int len, style, pos;
  QString sd;
  GHighlightStyle *st;
  bool italic = false;

  pos = 0;
  p.setFont(font());

  for (i = 0; i < l->highlight.count(); i ++)
  {
  	if (pos > (xmin + lmax))
  		break;

    style = l->highlight[i].state;
    len = l->highlight[i].len;
    st = &styles[style];

		if ((pos + len) >= xmin)
		{
			sd = l->s.mid(pos, len);

			/*if (style == GLine::Keyword)
			{
				if (getFlag(DrawWithRelief))
				{
					p.setPen(styles[GLine::Normal].color);
					p.drawText(x + 1, y + 1, sd);
					p.setPen(styles[style].color);
					p.drawText(x, y, sd);
				}
				else
				{
					p.setPen(styles[style].color);
					p.drawText(x, y, sd);
					p.drawText(x + 1, y, sd);
				}
			}
			else if (style == GLine::Error)
			{
				p.setPen(styles[GLine::Current].color);
				p.drawLine(x, y + 2, x + len * charWidth - 1, y + 2);
				p.setPen(styles[GLine::Normal].color);
				p.drawText(x, y, sd);
			}
			else
			{
				p.setPen(styles[style].color);
				p.drawText(x, y, sd);
			}*/

			if (st->italic != italic)
			{
				italic = st->italic;
				p.setFont(italic ? italicFont : font());
			}

			p.setPen(styles[style].color);
			p.drawText(x, y, sd);

			if (st->bold)
				p.drawText(x + 1, y, sd);

			if (st->underline)
				p.drawLine(x, y + 2, x + len * charWidth - 1, y + 2);
		}

    pos += len;
    x += len * charWidth;
  }

  if (pos < (int)l->s.length() && pos < (xmin + lmax))
  {
    p.setPen(styles[GLine::Normal].color);
    p.drawText(x, y, l->s.mid(pos));
  }
}



void GEditor::paintCell(QPainter * painter, int row, int)
{
  QRect ur = cellRect();
  GLine *l = doc->lines->at(row);
  QFontMetrics fm(painter->fontMetrics());
  int x1, y1, x2, y2;
  QColor color;
  int xmin, lmax;

  if (!l)
  {
    painter->fillRect(ur.left(), ur.top(), ur.width(), ur.height(), styles[GLine::Background].color);
    return;
  }

	xmin = (ur.left() - margin) / charWidth;
	lmax = 1 + ur.width() / charWidth;

  // Line background
  if (l->flag)
  {
    if (l->flag & GLine::CurrentFlag)
      color = styles[GLine::Current].color;
    else
      color = styles[GLine::Breakpoint].color;
  }
  else if (getFlag(ShowCurrentLine) && row == y)
    color = styles[GLine::Line].color;
  else
    color = styles[GLine::Background].color;

  cache->fill(color);

  QPainter p(cache);
  p.setFont(painter->font());
  p.translate(-ur.left(), -ur.top());

  // Selection background
  if (doc->hasSelection())
  {
    doc->getSelection(&y1, &x1, &y2, &x2);

    if (row >= y1 && row <= y2)
    {
      if (row > y1)
        x1 = 0;
      else
        x1 *= charWidth;

      if (row < y2)
        x2 = cellWidth() + 1;
      else
        x2 *= charWidth;

      p.fillRect(x1 + margin, 0, x2 - x1, cellHeight(), styles[GLine::Selection].color);
    }
  }

  // Margin
  if (margin)
  {
		if (!l->flag)
    	p.fillRect(0, 0, margin, cellHeight(), styles[GLine::Background].color);

    if (getFlag(ShowModifiedLines) && l->changed)
      p.fillRect(0, 0, margin - 1, cellHeight(), styles[GLine::Highlight].color);
    else if (getFlag(ShowLineNumbers))
      p.fillRect(0, 0, margin - 1, cellHeight(), styles[GLine::Line].color);

    x1 = 0;

    if (getFlag(ShowLineNumbers))
    {
      int n = row + 1;
      if ((n % 10) == 0)
      {
        p.setPen(styles[GLine::Normal].color);
        p.drawText(x1, fm.ascent(), QString::number(n).rightJustify(lineNumberLength));
      }
    }
  }

  // Highlight braces
  if (getFlag(HighlightBraces) && row == ym && x1m >= 0)
  {
    p.fillRect(x1m * charWidth + margin, 0, charWidth, cellHeight(), styles[GLine::Highlight].color);
    p.fillRect(x2m * charWidth + margin, 0, charWidth, cellHeight(), styles[GLine::Highlight].color);
  }

  // Line text
  if (doc->getHighlightMode() == GDocument::None || (l->modified && row == y && !getFlag(HighlightCurrent)) || l->flag)
  {
    p.setPen(styles[GLine::Normal].color);
    p.drawText(margin, fm.ascent(), l->s.mid(xmin, lmax));
  }
  else
  {
    doc->colorize(row);
    paintText(p, l, margin, fm.ascent(), xmin, lmax);
  }

  // Procedure separation lines (after colorize !)
  if (getFlag(ShowProcedureLimits) && l->proc)
  {
    QBrush brush(styles[GLine::Selection].color, Qt::Dense4Pattern);
    p.fillRect(0, 0, ur.width(), 1, brush);
  }

  // Text cursor
  if (cursor && row == y)
    p.fillRect(QMIN((int)l->s.length(), x) * charWidth + margin, 0, 2, cellHeight(), styles[GLine::Normal].color);

  p.end();

  painter->drawPixmap(ur.left(), ur.top(), *cache, 0, 0, ur.width(), ur.height());
}


void GEditor::checkMatching()
{
  static QString look("()[]{}");

  if (y < 0 || y >= (int)doc->lines->count())
    return;

  QString str = doc->lines->at(y)->s;
  int len = (int)str.length();

  QChar c, cr;
  int pos, d, l, bx, by, bx2;
  QChar s[len];
  bool ignore;
  int i;

  bx = -1;
  bx2 = -1;
  by = -1;
  ignore = false;

  if (len == 0)
    goto __OK;

  pos = -1;

  for (i = 0; i < len; i++)
  {
    c = str[i];

    if (ignore)
    {
      if (c == '\\')
      {
        s[i] = c;
        i++;
        s[i] = str[i];
        continue;
      }

      if (c == '"')
        ignore = false;
      else
        c = ' ';
    }
    else
    {
      if (c == '"')
        ignore = true;
    }

    s[i] = c;
  }

  if (x > 0)
  {
    bx2 = x - 1;
    c = s[bx2];
    pos = look.find(c);
  }

  if (pos < 0 && x <= len)
  {
    bx2 = x;
    c = s[bx2];
    pos = look.find(c);
  }

  if (pos >= 0)
  {
    d = (pos & 1) ? -1 : 1;
    cr = look[pos + d];

    pos = bx2;
    l = 0;
    for(;;)
    {
      pos += d;
      if (pos < 0 || pos >= len)
        break;

      if (s[pos] == c)
        l++;
      else if (s[pos] == cr)
      {
        l--;
        if (l < 0)
        {
          bx = pos;
          by = y;
          break;
        }
      }
    }
  }

__OK:

  if (by != ym || bx != x1m || bx2 != x2m)
  {
    if (ym >= 0)
      updateLine(ym);
    ym = by;
    x1m = bx;
    x2m = bx2;
    if (ym >= 0)
      updateLine(ym);
  }
}


bool GEditor::cursorGoto(int ny, int nx, bool mark)
{
  bool setxx;
  int px, py;

  if (!mark)
  {
    if (doc->hasSelection())
      doc->hideSelection();
  }

  setxx = xx != nx;

  if (ny < 0)
  {
    nx = QMAX(0, nx);
    ny = 0;
  }
  else if (ny > (numLines() - 1))
  {
    ny = numLines() - 1;
    nx = QMIN(nx, lineLength(ny));
  }
  else if (ny == y)
  {
    if (nx < 0 && ny > 0)
    {
      ny--;
      nx = lineLength(ny);
    }
    else if (nx > lineLength(ny) && ny < (numLines() - 1))
    {
      ny++;
      nx = 0;
    }
  }

  if (nx < 0)
    nx = 0;
  else if (nx > lineLength(ny))
    nx = lineLength(ny);

  if (mark)
  {
    if (!doc->hasSelection(this))
      doc->startSelection(this, y, x);
    doc->endSelection(ny, nx);
  }

  if (ny != y)
    doc->colorize(y);

  if (y != ny || x != nx)
  {
    updateLine(y);
    y = ny;
    x = nx;
    if (setxx)
      xx = x;
    updateLine(y);
    cursor = hasFocus();

    emit cursorMoved();
  }

  checkMatching();

  cursorToPos(y, x, &px, &py);
  if (px < margin || px >= (visibleWidth() - 2) || py < 0 || py >= (visibleHeight() - cellHeight() - 1))
    QApplication::postEvent(this, new QEvent(EVENT_ENSURE_VISIBLE));
  //ensureVisible(nx * charWidth, ny * cellHeight(), charWidth, cellHeight());

  return false;
}


void GEditor::insert(QString text)
{
  doc->eraseSelection();
  doc->insert(y, x, text);
  cursorGoto(doc->yAfter, doc->xAfter, false);
}

void GEditor::cursorLeft(bool shift, bool ctrl)
{
  if (ctrl && x > 0)
  {
    int nx;

    nx = doc->wordLeft(y, x, false);
    if (nx > 0)
      nx = doc->wordLeft(y, nx, true);

    cursorGoto(y, nx, shift);
  }
  else
    cursorGoto(y, x - 1, shift);
}


void GEditor::cursorRight(bool shift, bool ctrl)
{
  if (ctrl && x < lineLength(y))
  {
    int nx;

    nx = doc->wordRight(y, x, true);
    if (nx < lineLength(y))
      nx = doc->wordRight(y, nx, false);

    cursorGoto(y, nx, shift);
  }
  else
    cursorGoto(y, x + 1, shift);
}


void GEditor::cursorUp(bool shift, bool ctrl)
{
	if (!ctrl)
	{
  	cursorGoto(y - 1, xx, shift);
  	return;
	}

	int yy = y;
	for(;;)
	{
		yy--;
		if (yy <= 0)
			break;
		if (doc->lines->at(yy)->proc)
			break;
	}

	cursorGoto(yy, xx, shift);
}


void GEditor::cursorDown(bool shift, bool ctrl)
{
	if (!ctrl)
	{
  	cursorGoto(y + 1, xx, shift);
  	return;
	}

	int yy = y;
	for(;;)
	{
		yy++;
		if (yy >= numLines())
			break;
		if (doc->lines->at(yy)->proc)
			break;
	}

	cursorGoto(yy, xx, shift);
}

void GEditor::cursorHome(bool shift, bool ctrl)
{
  if (ctrl)
    cursorGoto(0, 0, shift);
  else
  {
		int indent = doc->getIndent(y);
		if (x == indent)
    	cursorGoto(y, 0, shift);
		else
    	cursorGoto(y, indent, shift);
	}
}


void GEditor::cursorEnd(bool shift, bool ctrl)
{
  if (ctrl)
    cursorGoto(numLines(), 0, shift);
  else
    cursorGoto(y, lineLength(y), shift);
}

void GEditor::cursorPageUp(bool mark)
{
  int page = visibleHeight() / cellHeight();
  cursorGoto(y - page, 0, mark);
}

void GEditor::cursorPageDown(bool mark)
{
  int page = visibleHeight() / cellHeight();
  cursorGoto(y + page, 0, mark);
}

void GEditor::newLine()
{
  doc->eraseSelection();
  doc->insert(y, x, "\n" + doc->lines->at(y)->s.left(QMIN(x, doc->getIndent(y))));
  cursorGoto(doc->yAfter, doc->xAfter, false);
}

void GEditor::del()
{
  if (doc->hasSelection())
  {
    doc->eraseSelection();
    return;
  }

  if (x == lineLength(y))
  {
    if (y < (numLines() - 1))
      doc->remove(y, x, y + 1, 0);
  }
  else
    doc->remove(y, x, y, x + 1);
}

void GEditor::backspace()
{
	int indent;
	int yy;

  if (doc->hasSelection())
  {
    doc->eraseSelection();
    return;
  }

  indent = doc->getIndent(y);

  if (x > 0 && x <= indent)
  {
  	yy = y;
  	indent = 0;
		while (yy > 0)
		{
			yy--;
			indent = doc->getIndent(yy);
			if (x > indent)
				break;
		}
		cursorGoto(y, indent, true);
		del();
  }
  else
  {
  	if (!cursorGoto(y, x - 1, false))
	    del();
	}
}

void GEditor::tab(bool back)
{
  QString ins;
  int y1, x1, y2, x2, yy;
  int indent, i;
  bool empty;
  int tw = doc->getTabWidth();

  if (!doc->hasSelection())
  {
    if (!back)
    {
      ins.fill(' ', tw - (x % tw));
      insert(ins);
    }
    return;
  }

  doc->getSelection(&y1, &x1, &y2, &x2);
  doc->startSelection(this, y1, 0);
  if (x2)
    y2++;
  doc->endSelection(y2, 0);

  indent = 65536;
  for (yy = y1; yy < y2; yy++)
  {
    i = doc->getIndent(yy, &empty);
    if (!empty)
      indent = QMIN(indent, i);
  }

  if (back && indent <= 0)
    return;

  doc->begin();

  if (!back)
  {
    ins.fill(' ', tw - (indent % tw));

    for (yy = y1; yy < y2; yy++)
    {
      doc->insert(yy, 0, ins);
      doc->colorize(yy);
      //if (r->s.stripWhiteSpace().length())
    }
  }
  else
  {
    indent %= tw;
    if (indent == 0)
      indent = tw;

    ins.fill(' ', indent);

    for (yy = y1; yy < y2; yy++)
    {
      if (doc->lines->at(yy)->s.left(indent) == ins)
      {
        doc->remove(yy, 0, yy, indent);
        doc->colorize(yy);
      }
    }
  }

  doc->end();
}


void GEditor::copy()
{
  doc->copy();
}

void GEditor::cut()
{
  doc->copy();
  doc->eraseSelection();
}

void GEditor::paste()
{
  QString text;
  QCString subType = "plain";
  uint i;
  QString tab;

  text = QApplication::clipboard()->text(subType);

  tab.fill(' ', doc->getTabWidth());
  text.replace("\t", tab);

  for (i = 0; i < text.length(); i++)
  {
    if ((text[i] < ' ' || text[i].isSpace()) && text[i] != '\n')
      text[i] = ' ';
  }

  insert(text);
}

void GEditor::undo()
{
  if (!doc->undo())
    cursorGoto(doc->yAfter, doc->xAfter, false);
}

void GEditor::redo()
{
  if (!doc->redo())
    cursorGoto(doc->yAfter, doc->xAfter, false);
}

void GEditor::selectAll()
{
  cursorGoto(0, 0, false);
  cursorGoto(numLines(), 0, true);
}

void GEditor::keyPressEvent(QKeyEvent *e)
{
  bool shift = e->state() & ShiftButton;
  bool ctrl = e->state() & ControlButton;

	e->accept();

  if (e->text().length()
      && e->key() != Key_Return
      && e->key() != Key_Enter
      && e->key() != Key_Delete
      && e->key() != Key_Backspace
      && (!e->ascii() || e->ascii() >= 32))
  {
    insert(e->text());
    return;
  }

  switch (e->key())
  {
    case Key_Left:
      cursorLeft(shift, ctrl); return;
    case Key_Right:
      cursorRight(shift, ctrl); return;
    case Key_Up:
      cursorUp(shift, ctrl); return;
    case Key_Down:
      cursorDown(shift, ctrl); return;
    case Key_Enter: case Key_Return:
      newLine(); return;
    case Key_Home:
      cursorHome(shift, ctrl); return;
    case Key_End:
      cursorEnd(shift, ctrl); return;
    case Key_Backspace:
      backspace(); return;
    case Key_Delete:
      del(); return;
    case Key_Prior:
      cursorPageUp(shift); return;
    case Key_Next:
      cursorPageDown(shift); return;
    case Key_Tab:
      tab(false); return;
    case Key_BackTab:
      tab(true); return;
  }

  if (!ctrl)
    goto __IGNORE;

  switch(e->key())
  {
    case Key_C:
      copy(); return;
    case Key_X:
      cut(); return;
    case Key_V:
      paste(); return;
    case Key_Z:
      undo(); return;
    case Key_Y:
      redo(); return;
    case Key_A:
      selectAll(); return;
  }

__IGNORE:

  e->ignore();
}


void GEditor::posToCursor(int px, int py, int *y, int *x)
{
  int nx, ny;

  ny = rowAt(contentsY() + py);
  if (ny < 0)
    ny = 0;
  else if (ny >= numLines())
    ny = numLines() - 1;

  nx = (contentsX() + px - margin) / charWidth;
  nx = QMAX(0, QMIN(nx, lineLength(ny)));

  *y = ny;
  *x = nx;
}

void GEditor::cursorToPos(int y, int x, int *px, int *py)
{
  int npx, npy;

  npy = y * cellHeight() - contentsY();
  npx = x * charWidth - contentsX() + margin;

  *py = npy;
  *px = npx;
}

void GEditor::mousePressEvent(QMouseEvent *e)
{
  bool shift = e->state() & ShiftButton;
  int nx, ny;
  //stopAutoScroll();
  //d->dnd_startpos = e->pos();

  if (e->button() != MidButton && e->button() != LeftButton)
    return;

  posToCursor(e->pos().x(), e->pos().y(), &ny, &nx);
  cursorGoto(ny, nx, shift);
}

void GEditor::mouseMoveEvent(QMouseEvent *e)
{
	if (!scrollTimer->isActive())
  	scrollTimer->start(50, false);
}

void GEditor::mouseDoubleClickEvent(QMouseEvent *e)
{
  if (e->button() == LeftButton && !(e->state() & ShiftButton))
  {
    int xl, xr;

    xl = doc->wordLeft(y, x, true);
    xr = doc->wordRight(y, x, true);

    if (xr > xl)
    {
      doc->hideSelection();
      cursorGoto(y, xl, false);
      cursorGoto(y, xr, true);
    }
  }
}

void GEditor::mouseReleaseEvent(QMouseEvent *e)
{
  scrollTimer->stop();
}


void GEditor::resizeEvent(QResizeEvent *e)
{
  QGridView::resizeEvent(e);
  baptizeVisible();
  updateLength();
}

bool GEditor::event(QEvent *e)
{
  if (e->type() == EVENT_ENSURE_VISIBLE)
  {
    ensureVisible(x * charWidth, y * cellHeight() + cellHeight() / 2, margin + 2, cellHeight());
    return true;
  }
  else if (e->type() == QEvent::KeyPress)
  {
		QKeyEvent *ke = (QKeyEvent *)e;
		if (ke->key() == Key_Tab || ke->key() == Key_BackTab)
		{
			keyPressEvent(ke);
			if (ke->isAccepted())
				return TRUE;
		}
  }

  return QGridView::event(e);
}


void GEditor::startBlink()
{
  blinkTimer->start(QApplication::cursorFlashTime() / 2, false);
  cursor = true;
}

void GEditor::stopBlink()
{
  blinkTimer->stop();
  cursor = false;
  updateLine(y);
}

void GEditor::blinkTimerTimeout()
{
  cursor = !cursor;
  updateLine(y);
}

void GEditor::focusInEvent(QFocusEvent *)
{
  startBlink();
}

void GEditor::focusOutEvent(QFocusEvent *)
{
  //if (doc->getHighlightMode() == GDocument::Gambas)
  //  doc->colorize(y);
  stopBlink();
}


void GEditor::scrollTimerTimeout()
{
  int ny, nx;
  QPoint pos = mapFromGlobal(QCursor::pos());

  posToCursor(pos.x(), pos.y(), &ny, &nx);
  cursorGoto(ny, nx, true);
}

void GEditor::setStyle(int index, GHighlightStyle *style)
{
  if (index < 0 || index >= GLine::NUM_STATE)
    return;

  styles[index] = *style;
  if (index == 0)
    setPaletteBackgroundColor(style->color);
  updateContents();
}

void GEditor::getStyle(int index, GHighlightStyle *style)
{
  if (index < 0 || index >= GLine::NUM_STATE)
  {
    *style = styles[GLine::Background];
    return;
  }

  *style = styles[index];
}

void GEditor::setFlag(int f, bool v)
{
  if (v)
    flags |= (1 << f);
  else
    flags &= ~(1 << f);

  updateMargin();
  updateContents();
}

void GEditor::updateMargin()
{
  int nm = 2, lnl = 0;

  if (getFlag(ShowLineNumbers))
  {
    int count = numLines();

    while (count)
    {
      nm += charWidth;
      lnl++;
      count /= 10;
    }
  }
  else if (getFlag(ShowModifiedLines))
    nm += 4;

  if (nm != margin)
  {
    margin = nm;
    lineNumberLength = lnl;
    updateContents();
  }
}

void GEditor::docTextChanged()
{
	emit textChanged();
}


void GEditor::baptizeVisible()
{
	doc->baptizeUntil(lastVisibleRow());
}

